from copy import deepcopy
import logging
from os import environ
import shlex
import shutil
import socket
import subprocess
from pathlib import Path
from datetime import datetime
from .tables import zabbix
#from shlex import quote as shquote


def quote(s):
    # good enough? shlex.quote alone is adding too many extras
    # must be readable, not immediately reuseable
    r = repr(s)
    if (
        len(r) - 2 != len(s)
        or any(ch in s for ch in " ()|")
    ):
        return r

    return s

def DPopen(*args, **kwargs):
    stderr = kwargs.get("stderr", subprocess.PIPE)
    try:
        logger = logging.getLogger()
        if logger.isEnabledFor(logging.DEBUG):
            stderr = None
    except Exception:
        pass

    kwargs["stderr"] = stderr

    return subprocess.Popen(*args, **kwargs)

def process_repr(cmd=None, env=None):
    if cmd is None:
        return "None"

    rargs = map(quote, cmd)

    str_env = " ".join((
        f"{key}={quote(value)}"
        for key, value
        in env.items()))

    # Good enough
    output = ""
    if str_env:
        output += str_env + " \\\n"

    output += " ".join((
        f"\\\n    {line}" if line.startswith("-") else line for line in rargs
    ))

    return output


class Command(object):
    def __init__(self, command, env_extra={}, **kwargs) -> None:
        self.env_extra = env_extra
        self.env = {**environ, **env_extra}
        self.command = tuple(map(str, command))

    def reprexec(self):
        rargs = map(quote, self.command)

        str_env = " ".join((
            f"{key}={quote(value)}"
            for key, value
            in self.env_extra.items()))

        # Good enough
        output = ""
        if str_env:
            output += str_env + " \\\n"

        output += " ".join((
            f"\\\n    {line}" if line.startswith("-") else line for line in rargs
        ))

        return output

    __repr__ = reprexec

    def exec(self, **kwargs):
        """
        Wrapper for subprocess.run.
        
        force 'text' output and returns 'stdout' as a tuple of lines
        where the last line is omitted if empty (it generally is).

        Return None on error (actual error, not the process retvalue)
        """
        if "text" not in kwargs:
            kwargs["text"] = True

        # allow overloading of stdin and stdout
        # stderr is used for logging in case of errors
        stdin  = subprocess.PIPE if "stdin"  not in kwargs else kwargs["stdin"]
        stdout = subprocess.PIPE if "stdout" not in kwargs else kwargs["stdout"]
        stderr = subprocess.PIPE
        try:
            logger = logging.getLogger()
            if logger.isEnabledFor(logging.DEBUG):
                stderr = None
        except Exception:
            pass

        try:
            result = subprocess.run(
                self.command, **kwargs,
                env=self.env,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                check=True,
            )
            out = result.stdout

        except subprocess.CalledProcessError as e:
            #logging.debug(e.stderr.rstrip())
            logging.debug(f"return code {e.returncode}")
            return None
        except FileNotFoundError:
            logging.critical(f"Command not found \"{self.command[0]}\"")
            return None

        if stdout == subprocess.PIPE:
            lines = tuple(map(str.strip, out.split("\n")))
            if lines:
                if lines[-1] == "":
                    return lines[:-1]
            return result.returncode, lines

        return result.returncode, []


class CurryCommand(object):
    def __init__(self, cmd, env, env_extra):
        merged_env = {**env, **env_extra}
        self.cmd, self.env, self.env_extra = (
            list(map(str, cmd)),
            merged_env,
            env_extra,
        )

    def exec(self, *args):
        command = self.cmd + list(args)
        return run(command, env=self.env)

    def reprexec(self, *args):
        rargs = map(quote, self.cmd + list(args))

        str_env = " ".join((
            f"{key}={quote(value)}"
            for key, value
            in self.env_extra.items()))

        # Good enough
        output = ""
        if str_env:
            output += str_env + " \\\n"

        output += " ".join((
            f"\\\n    {line}" if line.startswith("-") else line for line in rargs
        ))
        
        return output

    __repr__ = reprexec


def run(*args, **kwargs):
    """
    Wrapper for subprocess.run.
    
    force 'text' output and returns 'stdout' as a tuple of lines
    where the last line is omitted if empty (it generally is).

    Return None on error (actual error, not the process retvalue)
    """
    if "text" not in kwargs:
        kwargs["text"] = True

    stderr = subprocess.PIPE
    try:
        logger = logging.getLogger()
        if logger.isEnabledFor(logging.DEBUG):
            stderr = None
    except Exception:
        pass

    try:
        result = subprocess.run(
            *args, **kwargs,
            check=True,
            stdout=subprocess.PIPE,
            stderr=stderr)
        out = result.stdout

    except subprocess.CalledProcessError as e:
        #logging.debug(e.stderr.rstrip())
        logging.debug(f"return code {e.returncode}")
        return None
    except FileNotFoundError:
        logging.critical(f"Command not found \"{args[0]}\"")
        return None

    lines = tuple(map(str.strip, out.split("\n")))
    if lines:
        if lines[-1] == "":
            return lines[:-1]

    return lines


def check_binary(*names):
    """Checks whether 'names' are all valid commands in the current shell."""

    for name in names:
        if shutil.which(name) is None:
            return False
    return True

    #out = run(("command", "-v", *names))
    #if out is None:
    #    return False
    #return len(out) == len(names)


def try_find_sockets(search, port):
    """Try to locate available postgresql sockets."""
    if not check_binary("netstat"):
        return tuple()

    out = run(("netstat", "-lxn"))
    sockets = []
    for line in out:
        # not perfect but it works reasonably enough
        try:
            path = Path(line.split()[-1])
        except IndexError:
            pass
            folder = path.parent
            name = path.name
            if search not in folder or str(port) not in name:
                continue

            sockets.append(path)

    return tuple(sockets)


"""
def rlookup(ip):
    output = run(["dig", "+noall", "+answer", "-x", ip])
    if not output:
        return None

    parts = list(map(str.strip, output.split()))
    if len(parts) == 0:
        return None
    dn = parts[-1]

    subdomains = dn.split(".")
    if len(subdomains) == 0:
        return None

    return subdomains[0]
"""

def rlookup(ipaddr):
    try:
        response = socket.gethostbyaddr(ipaddr) # socket.getfqdn?
        host = repr(response[0])
        return host
    except Exception:
        return False


def build_compress_command(profile):
    algo, level, extra = profile

    extension = {"xz": ".xz", "gzip": ".gz", "bzip2": ".bz2"}

    env = {}
    ext = extension[algo]

    if check_binary(algo):
        cmd = (algo, f"-{level}", ) + extra
        pipe = cmd
        return env, ext, cmd, pipe

    if check_binary("7z"):
        cmd = ("7z", "a", f"-t{algo}", )
        pipe = ("7z", "a", f"-t{algo}", "-si", )
        return env, cmd, ext, pipe

    raise NotImplemented(f"Compression binary not available '{algo}")


def build_tar_command(profile):
    if not check_binary("tar"):
        raise NotImplemented("Missing tar command")

    algo, level, extra = profile

    if algo == "tar":
        return {}, ("tar", "-cf", ), ".tar"

    extension = {"xz": ".tar.xz", "gzip": ".tar.gz", "bzip2": ".tar.bz2"}
    env_map = {"xz": "XZ_OPT", "gzip": "GZIP", "bzip2": "BZIP2"}
    tar_map = {"xz": "J", "gzip": "z", "bzip2": "j"}

    compr_env_var = env_map[algo]
    compr_flags = " ".join((f"-{level}", ) + extra)
    tar_flag = tar_map[algo]

    env = {compr_env_var: compr_flags}
    ext = extension[algo]
    cmd = ("tar", f"-c{tar_flag}f", )

    return env, cmd, ext


def parse_zabbix_version(query_result):
    raw_version = query_result[0]
    major = int(raw_version[:-6])
    minor = int(raw_version[-6:-4])
    revision = int(raw_version[-4:])
    
    version = f"{major}.{minor}.{revision}"

    return version, (major, minor, revision)


def create_name(args):
    dt = datetime.now().strftime("%Y%m%d-%H%M%S")
    return f"zabbix_{args.host}_{dt}"


def preprocess_tables_lists(args, table_list):
    logging.debug(f"Table list: {table_list!r}")
    logging.verbose(f"Tables found: {len(table_list)}")

    tables = set(table_list)
    config = tables.intersection(zabbix.config)
    monitoring = tables.intersection(zabbix.monitoring)
    unknown = tables.difference(config, monitoring)

    logging.verbose(f"Config tables: {len(config)}")
    logging.verbose(f"Monitoring tables: {len(monitoring)}")
    logging.verbose(f"Unknown tables: {len(unknown)}")

    nodata, ignore, fail = [], [], []
    if args.monitoring == "nodata":
        nodata += monitoring

    if args.unknown == "nodata":
        nodata += unknown
    elif args.unknown == "ignore":
        ignore += unknown 
    elif args.unknown == "fail":
        fail += unknown

    return sorted(ignore), sorted(nodata), sorted(fail)


def pretty_log_args(args):
    """Print arguments via 'logging.info' in a readable way"""

    keys = args._keys

    str_args = ["Arguments:"]
    for key in keys:
        value = getattr(args, key, None)
        #if value is None:
        #    continue
        if key == "passwd" and value is not None:
            value = "[omissis]"

        str_args.append(f"    {key:<24}: {value}")

    #for key, value in dict_args.items():
    #    if key in keys or key == "scope":
    #        continue
    #    str_args.append(f"    {key:<24}: {value}")

    logging.verbose("\n".join(str_args))


import logging
from os import fdopen, environ
import tempfile
import shutil
import atexit
from pathlib import Path
from subprocess import PIPE

from .utils import (
    DPopen, build_compress_command, check_binary, parse_zabbix_version,
    preprocess_tables_lists, process_repr,
)


def backup_mysql(args):
    logging.verbose(f"DBMS: MySQL or MariaSql")
    if not check_binary("mysql", "mysqldump"):
        return 1, "Missing binaries: check 'mysql' and 'mysqldump' are available and in PATH"

    args.scope["env"] = {}
    
    # Phase 0: setup authentication
    _mysql_auth(args)

    # Phase 1: Fetch database version
    select_db_version = "SELECT optional FROM dbversion;"
    raw_version = _mysql_query(args, select_db_version, "zabbix version query")
    if raw_version is None:
        return 2, "Could not retrieve db version (see logs using --debug)"

    version, _ = parse_zabbix_version(raw_version)
    with open("zabbix_dbversion", "w") as fh:
        fh.writelines(["mysql\n", "{version}\n"])

    logging.info(f"Zabbix version: {version}")

    # Phase 2: Perform the actual backup

    # select and filter tables: done here and passed to _mysql_dump for simplicity
    table_list_query = (
        f"SELECT table_name FROM information_schema.tables "
        f"WHERE table_schema='{args.dbname}';")

    table_cmd = _mysql_query(args, table_list_query, "zabbix tables list query")
    if table_cmd is None:
        return 3, "Could not retrieve table list (see logs)"

    table_list = sorted(table_cmd)
    ignore, nodata, fail = preprocess_tables_lists(args, table_list)

    schema_ignores = []
    data_ignores = []
    if fail:
        return 4, f"Unknwon tables: aborting ({fail!r})"

    if nodata:
        for table in nodata:
            data_ignores += ["--ignore-table", f"{args.dbname}.{table}"]

    if ignore:
        for table in ignore:
            schema_ignores += ["--ignore-table", f"{args.dbname}.{table}"]
            data_ignores += ["--ignore-table", f"{args.dbname}.{table}"]

    # Phase 3: perform acqual dump

    # dump schemas
    schema_params = [
        "--opt", "--single-transaction", "--skip-lock-tables",
        "--no-data", "--routines", ]
    schema_outpath = Path("schemas_dump.sql")

    schema_status = _mysql_dump(
        args, schema_params, schema_ignores, schema_outpath,
        "Dump schema command", logging.info)

    if not schema_status:
        return 5, "Could not execute schema dump (see logs)"

    # dump data
    data_outpath = Path("data_dump.sql")
    data_params = [
        "--opt", "--single-transaction", "--skip-lock-tables",
        "--no-create-info", "--skip-extended-insert", "--skip-triggers",
    ]
    dump_status = _mysql_dump(
        args, data_params, data_ignores, data_outpath,
        "Dump data command", logging.info)

    if not dump_status:
        return 5, "Could not execute data dump (see logs)"

    return 0, "+OK"


def _mysql_auth(args):
    # Use provided loginfile and leave it untouched
    if args.loginfile is not None:
        pass

    # Create temporary mylogin.cnf file
    elif args.passwd is not None:
        logincnf_fd, logincnf_path = tempfile.mkstemp(prefix="logincnf_", text=True)
        logincnf_path = Path(logincnf_path)
        fh = fdopen(logincnf_fd, "w")

        logincnf_path.chmod(0o600)
        fh.writelines([
                f"[client]\n",
                f"password={args.passwd}\n",
            ])
        fh.close()

        abs_logincnf = logincnf_path.absolute()
        atexit.register(lambda: abs_logincnf.unlink())

        args.loginfile = logincnf_path

    if args.keeploginfile:
        shutil.copy(args.loginfile, "./mylogin.cnf")


def _mysql_query(args, query, description="query", log_func=logging.debug):
    dbname = args.dbname
    env_extra = args.scope["env"]
    
    # mysql command will be used to inspect the database
    query_cmd = [
        "mysql",
        "--login-path=client",
    ]

    if args.read_mysql_config:
        query_cmd += ["--defaults-file", args.mysql_config, ]

    if args.loginfile:
        query_cmd += [f"--defaults-extra-file={args.loginfile}", ]

    if args.sock is None:
        query_cmd += ["--host", args.host]
    else:
        query_cmd += ["--socket", args.sock]

    query_cmd += [
        "--user", args.user,
        "--port", args.port,
        "--database", dbname,
        "--skip-column-names",
        "--execute",
        query,
    ]

    query_cmd = tuple(map(str, query_cmd))
    query_env = {**environ, **env_extra}

    log_func(f"{description}: \n{process_repr(query_cmd, env_extra)}")

    exec_query = DPopen(query_cmd, env=query_env, stdout=PIPE, text=True)
    stdout, stderr = exec_query.communicate()

    if exec_query.returncode != 0:
        if stderr is not None:
            logging.fatal(stderr)
        return None

    return stdout.splitlines()


def _mysql_dump(
    args, params, ignoring, outpath, description="dump cmd", log_func=logging.debug
):
    dbname = args.dbname
    env_extra = args.scope["env"]

    # do we have a compression flag for mysql?
    compressor_profile = args.scope.get("mysqlcompression", None)

    # base command and pre-flags based on user selection
    dump_cmd = [
        "mysqldump",
    ]

    if args.read_mysql_config:
        dump_cmd += ["--defaults-file", args.mysql_config]

    if args.loginfile:
        dump_cmd += [f"--defaults-extra-file={args.loginfile}"]

    if args.sock is None:
        dump_cmd += ["--host", args.host]
    else:
        dump_cmd += ["--socket", args.sock]

    dump_cmd += [
        "--user", args.user,
        "--port", args.port,
        "--compression-algorithms", "zlib,zstd,uncompressed",
    ]

    if args.columns:
        dump_cmd += ["--complete-insert", "--quote-names", ]
    
    if args.verbosity in ("very", "debug"):
        dump_cmd += ["--verbose"]

    # parameters from outside (data or schema)
    dump_cmd += params

    # if there's no compression let mysqldump save the file
    if compressor_profile is None:
        dump_cmd += ["--result-file", outpath]
    
    # database to dump and tables exclusion
    dump_cmd += [dbname]
    dump_cmd += ignoring


    # finalize dump command
    dump_env = {**environ, **env_extra}
    dump_cmd = tuple(map(str, dump_cmd))
    log_func(f"{description}: \n{process_repr(dump_cmd, env_extra)}")

    # setup a compress command if needed
    if compressor_profile:
        env, _, ext, compressor = build_compress_command(compressor_profile)

        compr_env = {**environ, **env_extra, **env}
        compr_cmd = compressor + (outpath.name + ext, )
        log_func(f"{description} compression: \n{process_repr(compr_cmd, env)}")

    # don't execute if dry run is enabled
    if args.dry_run:
        return True

    # either run a simple dump or a 'dump | compress'
    dump = DPopen(dump_cmd, env=dump_env, stdout=PIPE, text=True)

    if not compressor_profile:
        dump.communicate()
        return dump.returncode == 0
    else:
        compress = DPopen(compr_cmd, env=compr_env, stdin=dump.stdout)
        compress.communicate()
        return compress.returncode == 0 



from pathlib import Path
import logging
from getpass import getpass
from .parser_defaults import PSqlArgs, MySqlArgs
import sys


__all__ = ["postprocess"]


def postprocess(args: PSqlArgs|MySqlArgs, user_args):
    """
    Adjust the arguments according to zabbix behaviour and user selection.

    See comments inside this function and _handle* functions.
    """
    parser = args.scope["parser"]
    dbms = args.scope["dbms"]

    # user_args: actual provided arguments
    # def_args: only defaults via argparse
    # local

    # The final arguments are stored in args directly, the precedence order is:
    #
    #   1. user provided
    #   2. from configuration files
    #   3. from configuration files defaults
    #   4. from the last shipped default configuration files
    #   5. from this script default
    #
    # I.e. CLI arguments are the most important.
    
    # Precedence lists in the form of {key: [values...]}
    # where local_args[key][0] is the selected final value

    # Hostname / Socket: special case
    if user_args.host == "-":
        user_args.host = ""

    # Check port limits
    if user_args.port is not None and not (1024 <= user_args.port <= 65535):
        raise parser.error(f"Port must be between 1024 and 65535: {user_args.port!r}")

    # Implicit read from zabbix config if a file is provided
    if user_args.zbx_config:
        user_args.read_zabbix_config = True

    if dbms == "mysql":
        # Implicit read from mysql config if a file is provided
        if user_args.mysql_config:
            user_args.read_mysql_config = True

    # Implicit save_files if files is provided
    if user_args.files:
        user_args.save_files = True
        user_args.files = Path(user_args.files)

    
    for key, value in vars(user_args).items():
        if value is not None:
            setattr(args, key, value)

    if args.read_zabbix_config:
        _handle_zabbix_conf(args, user_args)

    if args.rotate < 0:
        raise parser.error(f"Rotate must be 0 or positive: {args.rotate!r}")

    # Collapse verbosity to a single variable ('verbosity')
    _handle_verbosity(args)

    if dbms == "mysql":
        _handle_mysqlcompression(args)

    # Handle archiving and compression, set archive type and compression level
    _handle_archiving(args)

    # Checks whether the output directory is a directory or 
    # that it can be created (parent exists and is a directory)
    _handle_output(args)

    # Prompt for password if necessary (do it last to fail early on other arguments)
    if args.loginfile is None and args.passwd == "-":
        print("(echo disabled for password input)", file=sys.stderr)
        args.passwd = getpass("password: ")

    return args


def _parse_compression(parser, compr):
    # either a number (level), an algo (with default level 6)
    # or a combination of algo and number

    if compr in ("xz", "gzip", "bzip2"):
        algo = compr
        level = "6"
    elif ":" in compr:
        algo, level = compr.split(":")    
    else: # assume level format (checked below)
        algo = "gzip"
        level = compr

    extra = tuple()

    # extreme compression for xz
    if algo == "xz" and level.endswith("e"):
        level = level[:-1]
        extra += ("--extreme", )

    # check algorithm and level
    if algo not in ("xz", "gzip", "bzip2"):
        raise parser.error(f"Invalid compression algorithm: {compr!r}")

    if (not level.isdecimal() or
        len(level) != 1 or          # level is not a single decimal
        level == "0"                # 0 not allowed (only supported by xz)
    ):
        raise parser.error(f"Invalid/unsupported compression level: {compr!r}")

    return algo, level, extra


def _handle_mysqlcompression(args):
    parser = args.scope["parser"]
    
    if args.mysqlcompression == "-":
        args.scope["mysqlcompression"] = None
        return

    profile = _parse_compression(parser, args.mysqlcompression)
    args.scope["mysqlcompression"] = profile


def _handle_archiving(args):
    parser = args.scope["parser"]

    # Uncompressed folder 
    if args.archive == "-":
        args.scope["archive"] = None
        return

    if args.archive == "tar":
        args.scope["archive"] = ("tar", None, None, )
        return

    profile = _parse_compression(parser, args.archive)
    args.scope["archive"] = profile


def _handle_verbosity(args):
    """Set verbosity level."""
    logger = logging.getLogger()
    if args.quiet:
        args.verbosity = "quiet"
        logger.setLevel(logging.ERROR)
    elif args.very_verbose:
        args.verbosity = "very"
        logger.setLevel(logging.VERBOSE)
    elif args.debug:
        args.verbosity = "debug"
        logger.setLevel(logging.DEBUG)
    else:
        args.verbosity = "normal"
        logger.setLevel(logging.INFO)


def _handle_output(args):
    """Check whether the output directory is useable."""
    parser = args.scope["parser"]

    if ((args.outdir.exists() and not args.outdir.is_dir()) or
        (args.outdir.parent.exists() and not args.outdir.parent.is_dir())
    ):
        raise parser.error(f"Output directory: cannot create or use {args.outdir!r}")


def _handle_zabbix_conf(args, user_args):
    # args is a mix of user provided arguments and defaults
    # user_args has only values provided by the user
    # we need to update args if and only if the user has not
    # provided a value for something that comes from the zabbix
    # configuration file

    zconfig = _zabbix_try_read_config(args.zabbix_config)

    zbx_var_map = (
        ("DBHost", "host", str, ),
        ("DBPort", "port", int, ),
        ("DBName", "dbname", str, ),
        ("DBSchema", "schema", str, ),
        ("DBUser", "user", str, ),
        ("DBPassword", "passwd", str, ),
        ("DBSocket", "sock", Path, ),
    )

    zconfig = _map_clean_vars(zconfig, zbx_var_map)

    # For postgres prefer socket over host in zabbix configuration
    if args.scope["dbms"] == "psql":
        if "sock" in zconfig and "host" not in zconfig:
            zconfig["dbms"] = zconfig["sock"]
            del zconfig["sock"]

    # Copy value from the config if it isn't provided by the user  
    for key, value in zconfig.items():
        user_value = getattr(user_args, key, None)
        if user_value is None:
            setattr(args, key, value)


def _zabbix_try_read_config(path):
    """
    Try reading key value pairs from zabbix config file.
    
    Empty values are skipped silently.
    The defaults from the last shipped version are used as a base. Then are
    overriden by defaults from 'path' and, lastly, from the actual specified values.
    """
    config = []

    with path.open("r") as fh:
        for line in map(str.strip, fh):
            if line == "" or line.startswith("#"):
                continue
            
            key, eq, value = map(str.strip, line.partition("="))
            if eq == "=" and len(key) and len(value):
                config[key] = value

    return config


def _map_clean_vars(config, map_clean_var):
    """Helper function to merge configuration variables."""
    return dict((
        (new_name, type(config[key]))
        for key, new_name, type in map_clean_var
        if key in config
    ))


from os import environ
from pathlib import Path
from .utils import DPopen, run
from .utils import build_tar_command, process_repr
from shutil import rmtree
import logging


def parse_save_files(files):
    """
    Read a list of files and directories.
    
    One item per line.
    Spaces are trimmed and blank lines or lines starting with a # are ignored.
    """
    with open(files, "r") as fh:
        for line in fh:
            line = line.strip()
            if line.startswith("#"):
                continue
            yield Path(line)


def save_files(args):
    """
    Copy files and directories as per user arguments.
    
    Default file list: (module) zabbixbackup/assets/files.
    """
    if args.save_files is True:
        files = args.files
        if files == "-":
            files = Path(__file__).parent / "assets" / "files"

        copy_files(files, Path("host_root"))


def copy_files(save_files, base_dir):
    """
    Copy a list of files or directories in base_dir.

    Directory structure is replicated.
    """
    base_dir = base_dir.absolute()
    items = parse_save_files(save_files)

    for item in items:
        if not item.exists():
            logging.info(f"Filepath not found {item}, ignoring...")
            continue

        dest = base_dir / item.absolute().relative_to("/")
        if not dest.parent.exists():
            dest.parent.mkdir(parents=True)
            # TODO: copy permission on entire directory tree?

        # TODO: remove dependency on cp?
        if run(["cp", "-a", str(item), str(dest)]) is None:
            logging.warning(f"Cannot copy {item}, ignoring...")
        else:
            logging.info(f"Copying {item}")


def archive(archive_dir, args):
    """
    Create the actual archive file.

    Based on user arguments it will be compressed accordingly.
    """
    scope = args.scope
    profile = scope["archive"]

    if profile is not None:
        env, cmd, ext = build_tar_command(profile)
        name = archive_dir.name
        name_ext = name + ext
        tar_cmd = cmd + (name_ext, name, )
        tar_env = {**environ, **env}

        logging.debug(f"Archive command: \n{process_repr(tar_cmd, env)}\n")

        archive_exec = DPopen(tar_cmd, env=tar_env)
        archive_exec.communicate()

        if archive_exec.returncode == 0:
            logging.debug(f"Delete plain folder archive: {archive_dir}\n")
            rmtree(archive_dir)
        
        return Path(name_ext).absolute()

    # Leave as plain directory
    return Path(archive_dir).absolute()


import argparse
from pathlib import Path

from .parser_defaults import PSqlArgs, MySqlArgs
from .parser_post import postprocess


_description = "zabbix dump for {dbms} inspired and directly translated from..."


def parse(argv):
    """
    Arguments CLI parser.

    What's going on:
        'build_parser' choose the DBMS examining the first argument
        'build_sub_parser' create the actual parser to parse the rest of the arguments
    
    Arguments are then handled and adjusted according to zabbix behaviour
    anduser selection (see @postprocess function).
    
    The double call to 'parse_args' is for properly separating user provided values from
    script default values. In order to keep it working as expected *every* value in the
    parser must have a default (otherwise will prompt the help screen to the user
    regardless of the provided arguments).
    TODO: add credits in the help screen
    """

    main_parser = build_parser()

    main_args, subargv = main_parser(argv)
    dbms = main_args.dbms

    if dbms in ("psql", "pgsql"):
        args = PSqlArgs()
        args.scope["dbms"] = "psql"
    elif dbms in ("mysql", ):
        args = MySqlArgs()
        args.scope["dbms"] = "mysql"
        pass
    else:
        raise NotImplemented(f"DBMS {dbms} not supported")

    sub_parser = build_sub_parser(args)
    args.scope["parser"] = sub_parser

    temp_args = sub_parser.parse_args(subargv)
    
    try:
        _defaults = temp_args.__dict__
        _blanks = dict((key, None) for key in temp_args.__dict__.keys())

        sub_parser.set_defaults(**_blanks)
        user_args = sub_parser.parse_args(subargv)
        sub_parser.set_defaults(**_defaults)

    except Exception:
        raise ValueError(
            "Parse error: should never happen here."
            "Really.. something is wrong.")

    return postprocess(args, user_args)


def build_parser():
    """
    Create the parser according to user selection (first argument).
    """

    parser = argparse.ArgumentParser(
        "zabbixbackup",
        description=_description.format(dbms="postgres or mysql"))

    subparsers = parser.add_subparsers(title='DBMS', dest="dbms", required=True)

    subparsers.add_parser(
        'psql',
        aliases=["pgsql"],
        help="(see zabbixbackup psql --help)")

    subparsers.add_parser(
        'mysql',
        help="(see zabbixbackup mysql --help)")

    def _parser(argv):
        subargv = argv[0:1]

        if len(subargv) == 0:
            subargv = ["-h"]

        return parser.parse_args(subargv), argv[1:]

    return _parser


def build_sub_parser(args):
    """
    Create the parser according to user selection (by @build_parser).
    """
    dbms = args.scope["dbms"]

    parser = argparse.ArgumentParser(
        f"zabbixbackup {dbms}",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=_description.format(dbms=dbms))

    parser.add_argument(
        "-z", "--read-zabbix-config",
        help="try to read database host and credentials from Zabbix config. "
            "Implicit if `--zabbix-config` is set.",
        action="store_true",
        default=args.read_zabbix_config)

    parser.add_argument(
        "-Z", "--zabbix-config",
        help="Zabbix config file path. "
        "Implicit if `--read-zabbix-config` is set.",
        default=args.zabbix_config,
        type=Path,
        dest="zbx_config")

    if dbms == "mysql":
        parser.add_argument(
            "-c", "--read-mysql-config",
            help="Read database host and credentials from MySQL config file. "
                "Implicit if `--mysql-config` is set.",
            action="store_true",
            default=args.read_mysql_config)

        parser.add_argument(
            "-C", "--mysql-config",
            help="MySQL config file path. "
                "Implicit if `--read-mysql-config` is set.",
            default=args.mysql_config,
            type=Path)

    parser.add_argument(
        "-D", "--dry-run",
        help="Do not create the actual backup, only show dump commands. "
            "Be aware that the database will be queried for tables selection and "
            "temporary folders and files will be created.",
        default=args.dry_run,
        action="store_true")


    connection = parser.add_argument_group("connection options")

    _host_help = "hostname/IP of DBMS server, to specify a blank value pass '-'."
    if dbms == "psql":
        _host_help += " If host starts with a slash it's interpreted as a socket directory."
        _host_help += " Special rules might apply (see postgre online documentation for sockets)."

    connection.add_argument(
        "-H", "--host",
        help=_host_help,
        default=args.host)

    connection.add_argument(
        "-P", "--port",
        help="DBMS port.",
        default=args.port)

    if dbms == "mysql":
        connection.add_argument(
            "-S", "--socket",
            help="path to DBMS socket file. "
                "Alternative to specifying host.",
            dest="sock",
            default=args.sock)

    connection.add_argument(
        "-u", "--username",
        help="database login user.",
        default=args.user,
        dest="user")

    connection.add_argument(
        "-p", "--passwd",
        help="database login password (specify '-' for an interactive prompt).",
        default=args.passwd)

    login_file = ".pgpass" if dbms == "psql" else "mylogin.cnf" 
    connection.add_argument(
        "--keep-login-file",
        help=f"if a credential file is created ({login_file}) do not "
            "delete it on exit.",
        action="store_true",
        default=args.keeploginfile,
        dest="keeploginfile")


    connection.add_argument(
        "--login-file",
        help=f"use '{login_file}' file for the authentication.",
        default=args.loginfile,
        dest="loginfile")

    connection.add_argument(
        "-d", "--database",
        help="database name.",
        default=args.dbname,
        dest="dbname")

    if dbms == "psql":
        connection.add_argument(
            "-s", "--schema",
            help="database schema.",
            default=args.schema)

    connection.add_argument(
        "-n", "--reverse-lookup",
        help="(NOT IMPLEMENTED) perform a reverse lookup of the IP address for the host.",
        action="store_true",
        default=args.rlookup,
        dest="rlookup")


    dump = parser.add_argument_group("dump action options")

    dump.add_argument(
        "-U", "--unknown-action",
        help="action for unknown tables.",
        default=args.unknown,
        choices=("dump", "nodata", "ignore", "fail"),
        dest="unknown")

    dump.add_argument(
        "-M", "--monitoring-action",
        help="action for monitoring table",
        default=args.monitoring,
        choices=("dump", "nodata"),
        dest="monitoring")

    dump.add_argument(
        "-N", "--add-columns",
        help="add column names in INSERT clauses and quote them as needed.",
        default=args.columns,
        action="store_true",
        dest="columns")

    compression = parser.add_argument_group("dump level compression options")

    if dbms == "psql":
        compression.add_argument(
            "-x", "--pgcompression",
            help="passed as-is to pg_dump --compress, might be implied by format.",
            default=args.pgcompression)

        compression.add_argument(
            "-f", "--pgformat",
            help="dump format, will mandate the file output format.",
            choices={"plain", "custom", "directory", "tar"},
            default=args.pgformat)

    if dbms == "mysql":
        compression.add_argument(
            "--mysqlcompression",
            help="dump level compression. "
                "Available formats are xz, gzip and bzip2. Use ':<LEVEL>' to set a compression "
                "level. I.e. --archive xz:6. See documentation for the details.",
            default=args.mysqlcompression)

    files = parser.add_argument_group("configuration files")

    files.add_argument(
        "--save-files",
        help="save folders and other files (see --files).",
        default=args.save_files,
        action="store_true")

    files.add_argument(
        "--files",
        help="save folders and other files as listed in this file. "
            "One line per folder or file, non existant will be ignored. "
            "Directory structure is replicated (copied via 'cp').",
        default=args.files)

    output = parser.add_argument_group("output options")
    
    output.add_argument(
        "-a", "--archive",
        help="archive level compression. 'tar' to create a tar archive, "
            "'-' to leave the backup uncompressed as a folder. "
            "Other available formats are xz, gzip and bzip2. "
            "Use ':<LEVEL>' to set a compression level. I.e. --archive xz:6",
        default=args.archive)

    output.add_argument(
        "-o", "--outdir",
        help="save database dump to 'outdir'.",
        default=args.outdir,
        type=Path)

    output.add_argument(
        "-r", "--rotate",
        help="rotate backups while keeping up 'R' old backups."
            "Uses filename to match '0=keep everything'.",
        default=args.rotate,
        type=int)

    verbosity = parser.add_argument_group("verbosity")
    verbosity_group = verbosity.add_mutually_exclusive_group()
    # In case it is needed to change the default value for this group,
    # 'postprocess' must be modified accordingly (else clause is the default
    # during verbosity handling)
    # TODO: choose what to print and in which form
    verbosity_group.add_argument(
        "-q", "--quiet",
        help="don't print anything except unrecoverable errors.",
        action="store_true",
        default=args.quiet)

    verbosity_group.add_argument(
        "-v", "--verbose",
        help="print informations.",
        action="store_true",
        default=args.verbose)

    verbosity_group.add_argument(
        "-V", "--very-verbose",
        help="print even more informations.",
        action="store_true",
        default=args.very_verbose)

    verbosity_group.add_argument(
        "--debug",
        help="print everything.",
        action="store_true",
        default=args.debug)

    return parser



from pathlib import Path
from shutil import rmtree
import re
import logging
from .parser_defaults import PSqlArgs, MySqlArgs


re_cfg = re.compile(r"""
    zabbix_                             # suffix
    (?P<host>[^_]+?)_                   # host name
    (?P<year>[0-9]{4})                  # yearmonthday-hourminutesecond
    (?P<month>[0-9]{2})                 #
    (?P<day>[0-9]{2})-                  #
    (?P<hour>[0-9]{2})                  #
    (?P<minute>[0-9]{2})                #
    (?P<second>[0-9]{2})                #
    #(?P<version>([0-9][.])+?[0-9]+?)    # zabbix version and eol
    (?P<ext>([.]tar([.](gz|xz|bz2))?))? # extension (empty if plain folder)
""", re.VERBOSE)


def rotate(args: PSqlArgs|MySqlArgs):
    """
    Perform an archive rotation keeping the last 'args.n' archives.
    """
    n = args.rotate
    
    if n <= 0:
        return
    
    host = args.host

    # create a list of tuples in the form of [(datetime as int, folder)]
    # in order to being able to sort it naturally
    backups = []
    for archive in Path(".").iterdir():
        if match := re_cfg.fullmatch(archive.name):
            d = match.groupdict()
            if d["host"] == host:
                int_dt = int(
                    f"{d['year']}{d['month']}{d['day']}"
                    f"{d['hour']}{d['minute']}{d['second']}"
                )
                backups.append((int_dt, archive))

    backups = sorted(backups)
    remove, keep = backups[:-n], backups[-n:]

    logging.info("Rotate backups")
    logging.info(f"Found {len(backups)} backup/s")
    logging.info(f"Deleting {len(remove)} and keeping {len(keep)} backup/s")

    for _, item in remove:
        logging.verbose(f"    deleting backup '{item}'")
        if not args.dry_run:
            if item.is_file():
                item.unlink()
            else:
                rmtree(item)

    for _, item in keep:
        logging.debug(f"    keeping backup '{item}'")




if __name__ == "__main__":
    import os
    from sys import argv
    import logging
    from pathlib import Path
    from .parser import parse
    from .utils import create_name, pretty_log_args
    from .backup_postgre import backup_postgresql
    from .backup_mysql import backup_mysql
    from .archiver import save_files, archive
    from .rotation import rotate
    import atexit

    # Parse and preprocess cli arguments
    args = parse(argv[1:])
    scope = args.scope

    # TODO: rlookup here
    outdir = args.outdir
    abs_outdir = outdir.absolute()

    name = create_name(args)
    archive_dir = outdir / name
    abs_archive_dir = archive_dir.absolute()

    # Chdir into this backup directory 
    logging.debug(f"Create archive diretory and chdir to it: {archive_dir}")
    prev_cwd = Path(os.getcwd()).absolute()
    atexit.register(lambda: os.chdir(prev_cwd))
    archive_dir.mkdir()
    os.chdir(archive_dir)

    log_path = Path("dump.log")

    # Create log in the destination directory
    logging.debug(f"Log file: {log_path}")
    log_fh = log_path.open("w")
    logger = logging.getLogger()
    logger_handler = logging.StreamHandler(log_fh)
    logger.addHandler(logger_handler)

    logging.debug(f"Log file: {log_path}")

    # Pretty print arguments as being parsed and processed
    pretty_log_args(args)

    if scope["dbms"] == "psql":
        status, message = backup_postgresql(args)
    elif scope["dbms"] == "mysql":
        status, message = backup_mysql(args)

    # exit immediately if something went wrong
    if status != 0:
        logging.fatal(message)
        exit(status)

    save_files(args)

    # Detach file logger 
    logger.removeHandler(logger_handler)
    log_fh.close()

    # No file logging from here

    # From now on operate from backups diretory
    os.chdir(abs_outdir)
    # Archive, compress and move the backup to the final destination
    archive_path = archive(abs_archive_dir, args)

    # Rotate backups
    os.chdir(abs_outdir)
    rotate(args)

    print(archive_path)



from pathlib import Path
from dataclasses import dataclass, field

# Script default values
# (these are important to allow to separate user provided values during parsing)
# 'dest' name actions in parsers must be the same as in these dataclasses

@dataclass
class PSqlArgs(object):
    read_zabbix_config: bool = False
    zabbix_config: Path      = Path("/etc/zabbix/zabbix_server.conf")

    dry_run: bool            = False
    host: str                = "127.0.0.1"
    port: int                = 5432
    user: str                = "zabbix"
    passwd: str              = None
    keeploginfile: bool      = False
    loginfile: Path|None     = None
    dbname: str              = "zabbix"
    schema: str              = "public"
    rlookup: bool            = True

    save_files: bool         = False
    files: str|Path          = "-"

    unknown: str             = "ignore"
    monitoring: str          = "nodata"
    columns: bool            = False
    pgformat: str            = "custom"
    pgcompression: str       = None
    outdir: str|Path         = Path(".")
    rotate: int              = 0

    quiet: bool              = False
    verbose: bool            = True
    very_verbose: bool       = False
    debug: bool              = False

    verbosity: str           = None # automatically set during parser post process
    archive: str             = "-"
    scope: dict              = field(default_factory=dict)

PSqlArgs._keys = [
    "read_zabbix_config", "zabbix_config",
    "host", "port", "user", "passwd", "keeploginfile", "loginfile",
    "dbname", "schema", "rlookup",
    "save_files", "files",
    "unknown", "monitoring",
    "columns", "pgformat", "pgcompression",
    "rotate", "outdir", "archive",
    "dry_run", "verbosity",
]


@dataclass
class MySqlArgs(object):
    read_zabbix_config: bool = False
    zabbix_config: Path      = Path("/etc/zabbix/zabbix_server.conf")
    read_mysql_config: bool  = False
    mysql_config: str|Path   = "/etc/mysql/my.cnf"

    dry_run: bool            = False
    host: str                = "127.0.0.1"
    port: int                = 3306
    sock: Path|None          = None
    user: str                = "zabbix"
    passwd: str              = None
    keeploginfile: bool      = False
    loginfile: Path|None     = None
    dbname: str              = "zabbix"
    rlookup: bool            = True

    save_files: bool         = False
    files: str|Path          = "-"

    unknown: str             = "ignore"
    monitoring: str          = "nodata"
    columns: bool            = False
    mysqlcompression: str    = "gzip:6"
    outdir: str|Path         = Path(".")
    rotate: int              = 0

    quiet: bool              = False
    verbose: bool            = True
    very_verbose: bool       = False
    debug: bool              = False

    verbosity: str           = None # automatically set during parser post process
    archive: str             = "-"
    scope: dict            = field(default_factory=dict)

MySqlArgs._keys = [
    "read_zabbix_config", "zabbix_config", "read_mysql_config", "mysql_config",
    "host", "port", "sock", "user", "passwd", "keeploginfile", "loginfile",
    "dbname", "rlookup",
    "save_files", "files",
    "unknown", "monitoring",
    "columns", "mysqlcompression",
    "rotate", "outdir", "archive",
    "dry_run", "verbosity",
]


import logging
logging.basicConfig(format='[%(levelname)-8s] %(message)s', level=logging.DEBUG)
version = 1

# Shouldn't have done that.. it works.. I hope
logging.VERBOSE= 17
logging.addLevelName(17, "VERBOSE")
def _verbose(msg, *args, **kwargs):
    logging.log(logging.VERBOSE, msg, *args, **kwargs)
logging.verbose = _verbose
logger = logging.getLogger()


import logging
from os import fdopen, environ
import tempfile
import shutil
import atexit
from pathlib import Path
from subprocess import PIPE

from .utils import (
    DPopen, check_binary, parse_zabbix_version,
    preprocess_tables_lists, process_repr, try_find_sockets,
)


def backup_postgresql(args):
    logging.info(f"DBMS: Postgresql")
    if not check_binary("psql", "pg_dump"):
        return 1, "Missing binaries: check 'psql' and 'pg_dump' are available and in PATH"

    args.scope["env"] = {}
    
    # Phase 0: setup authentication
    _psql_auth(args)

    # Informational data about an eventual connection via socket
    if args.host == "" or args.host == "localhost" or args.host.startswith("/"):
        sockets = try_find_sockets("postgres", args.port)
        logging.info(f"sockets (actual choice performed directly by postgresql): ")
        logging.info(f"    {sockets!r}")


    # Phase 1: Fetch database version and assign a name
    select_db_version = "SELECT optional FROM dbversion;"
    raw_version = _psql_query(args, select_db_version, "zabbix version query")
    if raw_version is None:
        return 2, "Could not retrieve db version (see logs)"

    version, _ = parse_zabbix_version(raw_version)
    with open("zabbix_dbversion", "w") as fh:
        fh.writelines(["postgres\n", "{version}\n"])

    logging.info(f"Zabbix version: {version}")

    # Phase 2: Perform the actual backup
    
    # select and filter tables: done here and passed to _pg_dump for simplicity
    table_list_query = (
        f"SELECT table_name FROM information_schema.tables "
        f"WHERE table_schema='{args.schema}' AND "
        f"table_catalog='{args.dbname}' AND "
        f"table_type='BASE TABLE';")

    table_cmd = _psql_query(args, table_list_query, "zabbix tables list query")
    if table_cmd is None:
        return 3, "Could not retrieve table list (see logs)"

    table_list = sorted(table_cmd)
    ignore, nodata, fail = preprocess_tables_lists(args, table_list)

    dump_params = []

    if fail:
        return 4, f"Unknwon tables: aborting ({fail!r})"

    if nodata:
        for i in range(0, len(nodata), 4):
            nodata_pattern = f"({'|'.join(nodata[i:i+4])})"
            dump_params += ["--exclude-table-data", nodata_pattern]

    if ignore:
        for i in range(0, len(ignore), 4):
            ignore_pattern = f"({'|'.join(ignore)})"
            dump_params += ["--exclude-table", ignore_pattern]

    # all other flags and arguments are set up by _pg_dump
    outpath = Path("zabbix_dump")
    dump_status = _pg_dump(args, dump_params, outpath, "pgdump command", logging.info)
    if not dump_status:
        return 5, "Could not execute dump (see logs)"

    return 0, "+OK"


def _psql_auth(args):
    # Use provided loginfile and leave it untouched
    if args.loginfile is not None:
        args.scope["env"] = {"PGPASSFILE": str(args.loginfile)}

    elif args.passwd is not None:
        # Create temporary pgpass file
        pgpass_fd, pgpass_path = tempfile.mkstemp(prefix="pgpass_", text=True)
        pgpass_path = Path(pgpass_path)
        fh = fdopen(pgpass_fd, "w")

        pgpass_path.chmod(0o600)
        fh.writelines([
                # TODO: socket?
                f"{args.host}:{args.port}:{args.dbname}:{args.user}:{args.passwd}\n",
                "",
            ])
        fh.close()
        
        abs_pgpass = pgpass_path.absolute()
        atexit.register(lambda: abs_pgpass.unlink())

        args.loginfile = pgpass_path
        args.scope["env"] = {"PGPASSFILE": str(pgpass_path)}

    if args.keeploginfile:
        shutil.copy(args.loginfile, "./pgpass")
        Path("./pgpass")


def _psql_query(args, query, description="query", log_func=logging.debug):
    dbname = args.dbname
    env_extra = args.scope["env"]

    # psql command will be used to inspect the database
    query_cmd = [
        "psql",
        "--host", args.host,
        "--username", args.user,
        "--port", args.port,
        "--dbname", dbname,
        "--no-password",
        "--no-align",
        "--tuples-only",
        "--no-psqlrc",
        "--command",
        query,
    ]

    query_cmd = tuple(map(str, query_cmd))
    query_env = {**environ, **env_extra}

    log_func(f"{description}: \n{process_repr(query_cmd, env_extra)}")

    exec_query = DPopen(query_cmd, env=query_env, stdout=PIPE, text=True)
    stdout, stderr = exec_query.communicate()

    if exec_query.returncode != 0:
        if stderr is not None:
            logging.fatal(stderr)
        return None

    return stdout.splitlines()


def _pg_dump(
    args, params, outpath, description="dump cmd", log_func=logging.debug
):
    dbname = args.dbname
    env_extra = args.scope["env"]

    dump_cmd = [
        "pg_dump",
        "--host", args.host,
        "--username", args.user,
        "--port", args.port,
        "--dbname", dbname,
        "--schema", args.schema,
        "--no-password",
    ]

    if args.columns:
        # TODO: figure out if --inserts is redundant
        dump_cmd += ["--inserts", "--column-inserts", "--quote-all-identifiers", ]

    dump_cmd += ["--format", args.pgformat]

    if args.pgcompression is not None:
        dump_cmd += ["--compress", args.pgcompression]

    # choose the extension depending on output format
    extensions = {"plain": ".sql", "custom": ".pgdump", "directory": "", "tar": ".tar"}
    ext = extensions[args.pgformat]

    # try to guess the correct extension in case of dump compression,
    # good enough, might fail in some edge cases
    compr_ext = ""
    if args.pgcompression is not None and args.pgformat == "plain":
        algo, _, detail = args.pgcompression.partition(":")

        if algo == "0" or detail == "0":
            pass
        elif algo == "gzip" or algo.isdigit():
            compr_ext = ".gz"
        elif algo == "lz4":
            compr_ext = ".lz"
        elif algo == "zstd":
            compr_ext = ".zst"

    dump_path = f"{outpath}{ext}{compr_ext}"

    dump_cmd += ["--file", str(dump_path)]

    if args.verbosity in ("very", "debug"):
        dump_cmd += ["--verbose"]

    dump_cmd += params

    dump_env = {**environ, **env_extra}
    dump_cmd = tuple(map(str, dump_cmd))

    log_func(f"{description}: \n{process_repr(dump_cmd, env_extra)}")

    # don't execute if dry run is enabled
    if args.dry_run:
        return True

    dump = DPopen(dump_cmd, env=dump_env)
    dump.communicate()

    return dump.returncode == 0 



from types import SimpleNamespace as NS
from collections import namedtuple
from .tablesraw import raw_tables, CONFIG, MONITORING

table_type = namedtuple("TableSpec", ["name", "begin", "end", "data"])

# Table as a dict of TableSpec namedtuple in the form of
# 
#     {"tablename": TableSpec("name", "begin", "end", "data"), ...}
# 
# where begin and end are the first and last release the table has appeared in,
# data specify if the table is data or schemaonly flagged.

tables_spec = dict(
    (spec[0], table_type(*spec))
    for spec in raw_tables
)

all = set(tables_spec.keys())

config = set(
    name
    for name, spec in tables_spec.items()
    if spec.data == CONFIG
)

monitoring = set(
    name
    for name, spec in tables_spec.items()
    if spec.data == MONITORING
)

zabbix = NS(config=config, monitoring=monitoring, tables=all)